@@ -1,6 +1,7 @@
 (in-package :cl-user)
 
 (defparameter expanded '(let ((jss::this jss::*object-for-this*))
+  (declare (ignorable jss::this))
   (jcall "getLoaded"
          (jcall "load"
                 (jcall "make"
@@ -26,8 +27,6 @@
    .load(getClass().getClassLoader())
    .getLoaded()" )
 
-(in-package :jss-test)
-
 (prove:plan 1)
 (prove:is source expanded)
 
