@@ -84,6 +84,26 @@ public class JarPathname
     return result;
   }
 
+  /** Transform a reference to a nested Jar to an entry */
+  public static JarPathname archiveAsEntry(JarPathname p) {
+    JarPathname result = new JarPathname();
+    result = (JarPathname)result.copyFrom(p);
+
+    LispObject jars = result.getJars();
+    jars = jars.nreverse();
+    Pathname entry = (Pathname)jars.car();
+    jars = jars.cdr().nreverse();
+
+    result
+      .setDevice(jars)
+      .setDirectory(entry.getDirectory())
+      .setName(entry.getName())
+      .setType(entry.getType());
+    
+    return result;
+  }
+    
+
   /** Transform an entry in a jar to a reference as a jar */
   public static JarPathname createFromEntry(JarPathname p) {
     JarPathname result = new JarPathname();
@@ -91,20 +111,58 @@ public class JarPathname
       .copyFrom(p)
       .setDirectory(NIL)
       .setName(NIL)
-      .setType(NIL);
+      .setType(NIL)
+      .setVersion(Keyword.NEWEST);
     Pathname entryPath = p.getEntryPath();
     LispObject device = result.getDevice();
     device = device.reverse().push(entryPath).reverse();
     result.setDevice(device);
     return result;
   }
+
+  @DocString(name="as-jar-pathname-archive",
+             args="pathname",
+             returns="jar-pathname",
+             doc="Returns PATHNAME as a reference to a JAR-PATHNAME archive"
+             + "\n"
+             + "If PATHNAME names an ordinary file, the resulting JAR-PATHNAME addresses the"
+             + "file as an archive.  If PATHNAME names an entry in an archive, the resulting"
+             + "JAR-PATHNAME addresses that entry as a zip archive within that archive.")
+  private static final Primitive AS_JAR_PATHNAME_ARCHIVE = new pf_as_jar_pathname_archive();
+  private static class pf_as_jar_pathname_archive extends Primitive {
+    pf_as_jar_pathname_archive() {
+      super("as-jar-pathname-archive", PACKAGE_EXT, true);
+    }
+    @Override
+    public LispObject execute(LispObject arg) {
+      if (arg instanceof AbstractString) {
+        arg = coerceToPathname(arg);
+      }
+      if (arg instanceof JarPathname) {
+        return createFromEntry((JarPathname)arg);
+      } if (arg instanceof Pathname) {
+        return createFromPathname((Pathname)arg);
+      }
+      type_error(arg,
+                 list(Symbol.OR,
+                      Symbol.PATHNAME, Symbol.URL_PATHNAME,
+                      Symbol.JAR_PATHNAME));
+      return (LispObject)UNREACHED;
+    }
+  };
     
   static public JarPathname createFromFile(String s) {
-    return JarPathname.create(JAR_URI_PREFIX + "file:" + s + JAR_URI_SUFFIX);
+    JarPathname result
+      = JarPathname.create(JAR_URI_PREFIX + "file:" + s + JAR_URI_SUFFIX);
+    result.setVersion(Keyword.NEWEST);
+    return result;
   }
 
   static public JarPathname createEntryFromFile(String jar, String entry) {
-    return JarPathname.create(JAR_URI_PREFIX + "file:" + jar + JAR_URI_SUFFIX + entry);
+    JarPathname result
+      = JarPathname.create(JAR_URI_PREFIX + "file:" + jar + JAR_URI_SUFFIX + entry);
+    result.setVersion(Keyword.NEWEST);
+    return result;
   }
 
   static public JarPathname createEntryFromJar(JarPathname jar, String entry) {
@@ -122,7 +180,8 @@ public class JarPathname
     result
       .setDirectory(p.getDirectory())
       .setName(p.getName())
-      .setType(p.getType());
+      .setType(p.getType())
+      .setVersion(Keyword.NEWEST);
     
     return result;
   }
@@ -327,7 +386,7 @@ public class JarPathname
       
       Pathname withoutDevice = new Pathname();
       withoutDevice
-	.copyFrom(this)
+        .copyFrom(this)
         .setDevice(NIL);
 
       String withoutDeviceNamestring = withoutDevice.getNamestring(); // need to URI encode?
@@ -365,7 +424,7 @@ public class JarPathname
   }
 
   public static LispObject truename(Pathname pathname,
-				    boolean errorIfDoesNotExist) {
+                                    boolean errorIfDoesNotExist) {
     if (!(pathname instanceof JarPathname)) {
       return URLPathname.truename(pathname, errorIfDoesNotExist);
     }
@@ -392,7 +451,7 @@ public class JarPathname
       }
       LispObject rootJarTruename = Pathname.truename(rootJar, errorIfDoesNotExist);
       if (rootJarTruename.equals(NIL)) {
-	return Pathname.doTruenameExit(rootJar, errorIfDoesNotExist);
+        return Pathname.doTruenameExit(rootJar, errorIfDoesNotExist);
       }
       LispObject otherJars = p.getJars().cdr();
       URLPathname newRootJar;
