@@ -34,13 +34,15 @@
 package org.armedbear.lisp;
 
 import java.lang.ref.WeakReference;
+import java.lang.reflect.Method;
+import java.lang.reflect.InvocationTargetException;
 import static org.armedbear.lisp.Lisp.*;
 
 import java.util.Iterator;
 import java.util.concurrent.ConcurrentHashMap;
 import java.util.concurrent.ConcurrentLinkedQueue;
 import java.util.concurrent.atomic.AtomicInteger;
-
+import java.util.Stack; 
 import java.text.MessageFormat;
 
 public final class LispThread extends LispObject
@@ -79,12 +81,66 @@ public final class LispThread extends LispObject
     private Symbol wrapper =
         PACKAGE_THREADS.intern("THREAD-FUNCTION-WRAPPER");
 
+    /** Stack holding bindings for evaluated functions */
+    public Stack<Environment>envStack = new Stack<Environment>();
+
     LispThread(Thread javaThread)
     {
         this.javaThread = javaThread;
         name = new SimpleString(javaThread.getName());
     }
 
+  public static boolean virtualThreadingAvailable() {
+    try {
+      Class clazz = Class.forName("java.lang.Thread");
+      Class[] parameters = { java.lang.Runnable.class };
+      Method method = clazz.getDeclaredMethod("startVirtualThread", parameters);
+      return true;
+    } catch (ClassNotFoundException e1) {
+      Debug.trace("Failed to get java.lang.Thread by name");
+    } catch (NoSuchMethodException e2) {
+      // This is the case in non-Loom JVMs
+    } catch (SecurityException e3) {
+      Debug.trace("SecurityException caught introspecting threading interface: " + e3.toString());
+    }
+    return false;
+  }
+
+  public static Symbol NATIVE_THREADS = internKeyword("NATIVE");
+  public static Symbol VIRTUAL_THREADS = internKeyword("VIRTUAL");
+
+  static {
+    if (virtualThreadingAvailable()) {
+      Symbol._THREADING_MODEL.initializeSpecial(VIRTUAL_THREADS);
+    } else { 
+      Symbol._THREADING_MODEL.initializeSpecial(NATIVE_THREADS);
+    }
+  }
+
+  static Method threadBuilder = null;
+  static Method builderName = null;
+  static Method builderDaemon = null;
+  static Method builderVirtual = null;
+  static Method builderTask = null;
+  static Method builderBuild = null;
+
+  static {
+    try {
+      Class clazz = Class.forName("java.lang.Thread");
+      threadBuilder = clazz.getDeclaredMethod("builder");
+      clazz = Class.forName("java.lang.Thread$Builder");
+      builderDaemon = clazz.getDeclaredMethod("daemon", boolean.class);
+      builderName = clazz.getDeclaredMethod("name", String.class);
+      builderVirtual = clazz.getDeclaredMethod("virtual");
+      builderTask = clazz.getDeclaredMethod("task", java.lang.Runnable.class);
+      builderBuild = clazz.getDeclaredMethod("build");
+    } catch (Exception e) {
+      if (virtualThreadingAvailable()) {
+        Debug.trace("Failed to introspect virtual threading methods: " + e);
+      }
+    }
+  }
+
     LispThread(final Function fun, LispObject name)
     {
         Runnable r = new Runnable() {
@@ -116,13 +172,43 @@ public final class LispThread extends LispObject
                 }
             }
         };
-        javaThread = new Thread(r);
         this.name = name;
-        map.put(javaThread, this);
-        if (name != NIL)
-            javaThread.setName(name.getStringValue());
-        javaThread.setDaemon(true);
-        javaThread.start();
+
+        Thread thread = null;
+        
+        if (Symbol._THREADING_MODEL.getSymbolValue().equals(NATIVE_THREADS)) {
+          thread = new Thread(r);
+          if (name != NIL) {
+            thread.setName(name.getStringValue());
+          }
+          thread.setDaemon(true);
+        } else {
+          synchronized (threadBuilder) { // Thread.Builder isn't thread safe
+            Object o = null;
+            try {
+              o = threadBuilder.invoke(null);
+              if (name != NIL) {
+                o = builderName.invoke(o, name.getStringValue());
+              }
+              o = builderDaemon.invoke(o, true);
+              o = builderVirtual.invoke(o);
+              o = builderTask.invoke(o, r);
+              thread = (java.lang.Thread)builderBuild.invoke(o);
+            } catch (IllegalAccessException e1) {
+              Debug.trace("Use of reflection to start virtual thread failed: " + e1.toString());
+            } catch (InvocationTargetException e2) {
+              Debug.trace("Failed to invoke method to start virtual thread: " + e2.toString());
+            }
+          }
+        }
+        if (thread == null) {
+          Debug.trace("Failed to create java.lang.Thread");
+          javaThread = null;
+        } else {
+          javaThread = thread;
+          map.put(javaThread, this);
+          javaThread.start();
+        }
     }
 
     public StackTraceElement[] getJavaStackTrace() {
@@ -454,7 +540,7 @@ public final class LispThread extends LispObject
                 thread.specials[index] = null;
             }
 
-            freeSpecialIndices.add(new Integer(index));
+            freeSpecialIndices.add(Integer.valueOf(index));
         }
     }
 
@@ -721,6 +807,7 @@ public final class LispThread extends LispObject
                 StackMarker marker = (StackMarker) stackObj;
                 int numArgs = marker.getNumArgs();
                 LispStackFrame frame = new LispStackFrame(stk, framePos - numArgs - STACK_FRAME_EXTRA, numArgs);
+                frame.thread = this;
                 stk[framePos - 1] = frame;
                 if (prev != null) {
                     prev.setNext(frame);
@@ -795,9 +882,11 @@ public final class LispThread extends LispObject
         stack[stackPtr + 1] = STACK_MARKER_0;
         stackPtr += STACK_FRAME_EXTRA;
         try {
+            envStack.push(new Environment(null,NIL,function));
             return function.execute();
         }
         finally {
+            envStack.pop() ; 
             popStackFrame(0);
         }
     }
@@ -811,9 +900,11 @@ public final class LispThread extends LispObject
         stack[stackPtr + 2] = STACK_MARKER_1;
         stackPtr += 1 + STACK_FRAME_EXTRA;
         try {
+            envStack.push(new Environment(null,NIL,function));
             return function.execute(arg);
         }
         finally {
+            envStack.pop() ; 
             popStackFrame(1);
         }
     }
@@ -829,9 +920,11 @@ public final class LispThread extends LispObject
         stack[stackPtr + 3] = STACK_MARKER_2;
         stackPtr += 2 + STACK_FRAME_EXTRA;
         try {
+            envStack.push(new Environment(null,NIL,function));
             return function.execute(first, second);
         }
         finally {
+            envStack.pop() ; 
             popStackFrame(2);
         }
     }
@@ -848,9 +941,11 @@ public final class LispThread extends LispObject
         stack[stackPtr + 4] = STACK_MARKER_3;
         stackPtr += 3 + STACK_FRAME_EXTRA;
         try {
+            envStack.push(new Environment(null,NIL,function));
             return function.execute(first, second, third);
         }
         finally {
+            envStack.pop() ; 
             popStackFrame(3);
         }
     }
@@ -869,9 +964,11 @@ public final class LispThread extends LispObject
         stack[stackPtr + 5] = STACK_MARKER_4;
         stackPtr += 4 + STACK_FRAME_EXTRA;
         try {
+            envStack.push(new Environment(null,NIL,function));
             return function.execute(first, second, third, fourth);
         }
         finally {
+            envStack.pop() ; 
             popStackFrame(4);
         }
     }
@@ -891,9 +988,11 @@ public final class LispThread extends LispObject
         stack[stackPtr + 6] = STACK_MARKER_5;
         stackPtr += 5 + STACK_FRAME_EXTRA;
         try {
+            envStack.push(new Environment(null,NIL,function));
             return function.execute(first, second, third, fourth, fifth);
         }
         finally {
+            envStack.pop() ; 
             popStackFrame(5);
         }
     }
@@ -915,9 +1014,11 @@ public final class LispThread extends LispObject
         stack[stackPtr + 7] = STACK_MARKER_6;
         stackPtr += 6 + STACK_FRAME_EXTRA;
         try {
+            envStack.push(new Environment(null,NIL,function));
             return function.execute(first, second, third, fourth, fifth, sixth);
         }
         finally {
+            envStack.pop() ; 
             popStackFrame(6);
         }
     }
@@ -940,10 +1041,12 @@ public final class LispThread extends LispObject
         stack[stackPtr + 8] = STACK_MARKER_7;
         stackPtr += 7 + STACK_FRAME_EXTRA;
         try {
+            envStack.push(new Environment(null,NIL,function));
             return function.execute(first, second, third, fourth, fifth, sixth,
                                     seventh);
         }
         finally {
+            envStack.pop();
             popStackFrame(7);
         }
     }
@@ -967,10 +1070,12 @@ public final class LispThread extends LispObject
         stack[stackPtr + 9] = STACK_MARKER_8;
         stackPtr += 8 + STACK_FRAME_EXTRA;
         try {
+            envStack.push(new Environment(null,NIL,function));
             return function.execute(first, second, third, fourth, fifth, sixth,
                                     seventh, eighth);
         }
         finally {
+            envStack.pop() ; 
             popStackFrame(8);
         }
     }
@@ -983,9 +1088,11 @@ public final class LispThread extends LispObject
         stack[stackPtr + args.length + 1] = new StackMarker(args.length);
         stackPtr += args.length + STACK_FRAME_EXTRA;
         try {
+            envStack.push(new Environment(null,NIL,function));
             return function.execute(args);
         }
         finally {
+            envStack.pop() ; 
             popStackFrame(args.length);
         }
     }
@@ -1368,6 +1475,7 @@ public final class LispThread extends LispObject
             for (int i = args.length; i-- > 2;)
                 funArgs = new Cons(args[i], funArgs);
             thread.interrupt(fun, funArgs);
+            setInterrupted(thread,true);
             return T;
         }
     };
