@@ -296,6 +296,8 @@ public class Pathname extends LispObject
       result.setDirectory(new Cons(Keyword.RELATIVE));
       return result;
     } 
+    if (s.startsWith("./"))
+      { s = s.substring(2); }
     if (s.equals("..") || s.equals("../")) {
       result.setDirectory(list(Keyword.RELATIVE, Keyword.UP));
       return result;
@@ -1293,8 +1295,14 @@ public class Pathname extends LispObject
             rootDevice = URLPathname.create(root);
           }
           result.setDevice(new Cons(rootDevice, result.getDevice().cdr()));
-          // sanity check that the pathname has been constructed correctly
 
+          if (result.getDirectory().equals(NIL)
+              && (!result.getName().equals(NIL)
+                  || !result.getType().equals(NIL))) {
+            result.setDirectory(NIL.push(Keyword.ABSOLUTE));
+          }
+
+          // sanity check that the pathname has been constructed correctly
           result.validateComponents();
           return result;
         }
@@ -1740,7 +1748,7 @@ public class Pathname extends LispObject
           result.setDevice(p.getDevice());
         } else {
           if (d instanceof JarPathname
-	      && p instanceof JarPathname) {
+              && p instanceof JarPathname) {
             result.setDevice(d.getDevice());
           } else {
             result.setDevice(p.getDevice());
@@ -1760,11 +1768,11 @@ public class Pathname extends LispObject
             result.setDevice(d.getDevice());
           }
         } else {
-	  if (p.isLocalFile()) {
-	    result.setDevice(d.getDevice());
-	  } else {
-	    result.setDevice(p.getDevice());
-	  }
+          if (p.isLocalFile()) {
+            result.setDevice(d.getDevice());
+          } else {
+            result.setDevice(p.getDevice());
+          }
         }
       }
 
@@ -1793,6 +1801,15 @@ public class Pathname extends LispObject
       } else {
         result.setType(d.getType());
       }
+
+      // JAR-PATHNAME directories are always absolute
+      if ((result instanceof JarPathname)
+          && (!result.getName().equals(NIL)
+              || !result.getType().equals(NIL))
+          && result.getDirectory().equals(NIL)) {
+        result.setDirectory(NIL.push(Keyword.ABSOLUTE));
+      }
+            
       //  CLtLv2 MERGE-PATHNAMES 
     
       // "[T]he missing components in the given pathname are filled
